"""
empowering.service
~~~~~~~~~~~~~~

:copyrigth: (c) 2013 by GISCE-TI, S.L. See AUTHORS for more details.
:licese: EUPL, see LICENSE for more details.
"""


import json
from warnings import warn

from libsaas.services import base
from libsaas import http, parsers
from libsaas.executors import urllib2_executor as base_executor
from empowering.executors import urllib2_executor
from empowering.executors.urllib2_executor import (
    HTTPSClientAuthHandler, HTTPEmpoweringFilterHandler, HTTPAuthEmpowering,
)
from empowering.resource import EmpoweringResource
from empowering.results import *


class Contracts(EmpoweringResource):
    path = 'contracts'


class AmonMeasures(EmpoweringResource):
    path = 'amon_measures'

    @base.apimethod
    def get(self):
        raise base.MethodNotSupported

    @base.apimethod
    def update(self, obj, etag):
        raise base.MethodNotSupported

    @base.apimethod
    def delete(self, obj, etag):
        raise base.MethodNotSupported


class AmonMeasuresMeasurements(EmpoweringResource):
    path = 'amon_measures_measurements'

    @base.apimethod
    def create(self, obj):
        raise base.MethodNotSupported

    @base.apimethod
    def update(self, obj, etag):
        raise base.MethodNotSupported

    @base.apimethod
    def delete(self, start=None, end=None):
        params = base.get_params(('start', 'end'), locals())
        request = http.Request('DELETE', self.get_url())
        return request, parsers.parse_json


class Measures(AmonMeasuresMeasurements):
    def __init__(self, *args, **kwargs):
        super(AmonMeasuresMeasurements, self).__init__(*args, **kwargs)
        warn('Deprecated class. Use AmonMeasuresMeasurements',
             DeprecationWarning)


class Empowering(base.Resource):
    """
    Empowering Insight Engine Service API.
    """
    def __init__(self, company_id, username=None, password=None, key_file=None,
                 cert_file=None, version='v1', debug=False):
        self.company_id = str(company_id)
        self.key_file = key_file
        self.cert_file = cert_file
        self.version = version
        self.apiroot = "https://api.empowering.cimne.com"
        if debug:
            self.apiroot = "https://37.59.27.175"
        self.login_handler = None
        self.add_filter(self.use_json)
        self.add_filter(self.add_company_id)
        self.add_filter(self.add_cookie_token)
        self.setup_executor()
        if username and password:
            self.login(username, password)

    @property
    def token(self):
        return self.login_handler.token

    def setup_executor(self, extra_handlers=None):
        if extra_handlers is None:
            extra_handlers = ()
        extra_handlers += (HTTPEmpoweringFilterHandler(), )
        if self.key_file and self.cert_file:
            extra_handlers += (
                HTTPSClientAuthHandler(self.key_file, self.cert_file),
            )

        urllib2_executor.use(extra_handlers)

    def use_json(self, request):
        if request.method.upper() not in http.URLENCODE_METHODS:
            request.headers['Content-Type'] = 'application/json'
            request.params = json.dumps(request.params)

    def add_company_id(self, request):
        request.headers['X-CompanyId'] = self.company_id

    def add_cookie_token(self, request):
        if self.token:
            request.headers['Cookie'] = "iPlanetDirectoryPro=%s" % self.token

    def get_url(self):
        return "{0}/{1}".format(self.apiroot, self.version)

    def login(self, user, password):
        if self.login_handler and self.token:
            return {'success': True, 'token': self.token}
        endpoint = "{}/authn/login".format(self.apiroot)
        self.login_handler = HTTPAuthEmpowering(user, password, endpoint)
        auth = self.login_handler.login()
        self.setup_executor((self.login_handler, ))
        return auth

    def logout(self):
        if not self.token:
            return {'success': True}
        else:
            endpoint = '{0}/authn/logout'.format(self.apiroot)
            request = http.Request('GET', endpoint)
            self.use_json(request)
            self.add_company_id(request)
            self.add_cookie_token(request)

            executor = base_executor.base.current_executor()
            auth = executor(request, parsers.parse_json)
            if auth.get('success'):
                self.login_handler.token = None
            return auth

    @base.resource(AmonMeasures)
    def amon_measures(self):
        return AmonMeasures(self)

    @base.resource(Measures)
    def measures(self):
        warn('Deprecated. Use amon_measures_measurements', DeprecationWarning)
        return Measures(self)

    @base.resource(AmonMeasuresMeasurements)
    def amon_measures_measurements(self):
        return AmonMeasuresMeasurements(self)

    @base.resource(Contracts)
    def contract(self, contract_id):
        return Contracts(self, contract_id)

    @base.resource(Contracts)
    def contracts(self):
        return Contracts(self)

    @base.resource(OT101Results)
    def ot101_result(self, result_id):
        return OT101Results(self, result_id)

    @base.resource(OT101Results)
    def ot101_results(self):
        return OT101Results(self)

    @base.resource(OT103Results)
    def ot103_result(self, result_id):
        return OT103Results(self, result_id)

    @base.resource(OT103Results)
    def ot103_results(self):
        return OT103Results(self)

    @base.resource(OT106Results)
    def ot106_result(self, result_id):
        return OT106Results(self, result_id)

    @base.resource(OT106Results)
    def ot106_results(self):
        return OT106Results(self)

    @base.resource(OT201Results)
    def ot201_result(self, result_id):
        return OT201Results(self, result_id)

    @base.resource(OT201Results)
    def ot201_results(self):
        return OT201Results(self)

    @base.resource(OT204Results)
    def ot204_result(self, result_id):
        return OT204Results(self, result_id)

    @base.resource(OT204Results)
    def ot204_results(self):
        return OT204Results(self)

    @base.resource(OT503Results)
    def ot503_results(self):
        return OT503Results(self)

    @base.resource(OT900Results)
    def ot900_results(self):
        return OT900Results(self)

    @base.resource(BT111Results)
    def bt111_result(self, result_id):
        return BT111Results(self, result_id)

    @base.resource(BT111Results)
    def bt111_results(self):
        return BT111Results(self)



# -*- coding: utf-8 -*-

from libsaas.services import base
from libsaas import http, parsers

class EmpoweringResource(base.RESTResource):

    @base.apimethod
    def update(self, obj, etag):
        self.require_item()
        request = http.Request('PATCH', self.get_url(), self.wrap_object(obj),
                               headers={"If-Match": etag})
        return request, parsers.parse_json

    @base.apimethod
    def delete(self, etag):
        request = http.Request('DELETE', self.get_url(),
                               headers={"If-Match": etag})
        return request, parsers.parse_json

    @base.apimethod
    def get(self, where=None, sort=None):
        sort = sort and sort.replace(' ', '')
        params = base.get_params(('where', 'sort'), locals())
        request = http.Request('GET', self.get_url(), params)

        return request, parsers.parse_json

    def multiget(self, where=None, sort=None, max_results=None):
        query = where
        if max_results:
            query += '&max_results=%d' % max_results
        page = 0
        all_results = {'_items': []}
        more_items = True
        while more_items:
            more_items = False

            if page:
                paged_query = query + '&page=%s' % page
            else:
                paged_query = query
            result = self.get(where=paged_query, sort=sort)
            if '_items' in result:
                all_results['_items'].extend(result['_items'])

            if 'next' in result['_links']:
                page = result['_links']['next']['href'].split("&page=")[1]
                more_items = True
        return all_results





from datetime import datetime
import uuid
import times
from arrow.parser import DateTimeParser


# Monkey patch parse function for times library (newer)
parser = DateTimeParser()
times.parse = parser.parse_iso


def remove_none(struct, context=None):
    if not context:
        context = {}
    if 'xmlrpc' in context:
        return struct
    converted = struct.copy()
    for key, value in struct.items():
        if isinstance(value, dict):
            converted[key] = remove_none(value)
        else:
            if value is None or (isinstance(value, bool) and not value):
                del converted[key]
    return converted


def false_to_none(struct, context=None):
    if not context:
        context = {}
    if 'xmlrpc' in context:
        return struct
    converted = struct.copy()
    for key, value in struct.items():
        if isinstance(value, dict):
            converted[key] = false_to_none(value)
        else:
            if isinstance(value, bool) and not value:
                converted[key] = None
    return converted


def none_to_false(struct):
    if isinstance(struct, (list, tuple)):
        return [none_to_false(x) for x in struct]
    converted = struct.copy()
    for key, value in struct.items():
        if isinstance(value, dict):
            converted[key] = none_to_false(value)
        elif isinstance(value, (list, tuple)):
            converted[key] = none_to_false(value)
        else:
            if value is None:
                converted[key] = False
    return converted


def make_uuid(model, model_id):
    if isinstance(model, unicode):
        model = model.encode('utf-8')
    if isinstance(model_id, unicode):
        model_id = model_id.encode('utf-8')
    token = '%s,%s' % (model, model_id)
    return str(uuid.uuid5(uuid.NAMESPACE_OID, token))


def make_utc_timestamp(timestamp, timezone='Europe/Madrid'):
    if not timestamp:
        return None
    return times.to_universal(timestamp, timezone).isoformat('T') + 'Z'


def make_local_timestamp(timestamp, timezone='Europe/Madrid'):
    if not timestamp:
        return None
    if isinstance(timestamp, basestring):
        timestamp = times.parse(timestamp.replace('Z', ''))
    return times.to_local(timestamp, timezone).strftime('%Y-%m-%d %H:%M:%S')


def datestring_to_epoch(date_string):
    if not date_string:
        return None
    if not isinstance(date_string, datetime):
        dt = datetime.strptime(date_string, '%Y-%m-%d')
    else:
        dt = date_string
    return dt.strftime('%s')

def searchparams_to_querystring(search_params):
    operator_map = {
        '=': '==',
        '>=': '>=',
        '>': '>',
        '<': '<',
        '<=': '<=',
    }

    query = ''
    for param in search_params:
        if query:
            # is not the first
            query += ' and '

        field = param[0]
        operator = param[1]
        value = param[2]

        try:
            query_operator = operator_map[operator]
        except KeyError:
            raise Exception('Unsuported operand "%s"' % operand)

        if type(value) is str:
            # Add "" to the value
            query_value = '"%s"' % value
        else:
            query_value = '%s' % value

        query += '"%s"%s%s' % (field, query_operator, query_value)

    return query




import urllib2
import httplib
import logging
import json
from urlparse import urlparse, urlunparse
from libsaas.executors import base, urllib2_executor


logger = logging.getLogger('empowering.executors.urllib2_executor')


class HTTPSClientAuthHandler(urllib2.HTTPSHandler):
    """HTTPS Client Auth Handler.

    (c) Kalys Osmonov - http://www.osmonov.com/2009/04/client-certificates-with-urllib2.html
    """
    def __init__(self, key_file, cert_file):
        urllib2.HTTPSHandler.__init__(self)
        self.key_file = key_file
        self.cert_file = cert_file

    def https_open(self, req):
        # Rather than pass in a reference to a connection class, we pass in
        # a reference to a function which, for all intents and purposes,
        # will behave as a constructor
        return self.do_open(self.getConnection, req)

    def getConnection(self, host, timeout=300):
        logger.debug('using https connection (key file:{} Cert file:{})'.format(
            self.key_file, self.cert_file
        ))
        return httplib.HTTPSConnection(host, key_file=self.key_file,
                                       cert_file=self.cert_file)


class HTTPEmpoweringFilterHandler(urllib2.BaseHandler):

    def http_request(self, request):
        if request.get_method() == 'GET':
            url = urlparse(request.get_full_url())
            newurl = urlunparse((
                url.scheme,
                url.netloc,
                url.path,
                url.params,
                urllib2.unquote(url.query),
                url.fragment
            ))
            logger.debug("New url set to %s" % newurl)
            newr = urllib2_executor.RequestWithMethod(
                newurl, request.data, request.headers, request.origin_req_host,
                request.unverifiable
            )
            newr.timeout = request.timeout
            newr.set_method(request.get_method())
            return newr
        else:
            return request
    https_request = http_request


class HTTPAuthEmpowering(urllib2.BaseHandler):

    def __init__(self, username, password, endpoint):
        self.username = username
        self.password = password
        self.token = None
        self.endpoint = endpoint
        self.retried = 0
        self.max_retries = 1

    def http_error_401(self, req, fp, code, msg, headers):
        logger.debug("Login required. Retry auth")
        response = self.retry_auth(req, headers)
        self.reset_retry_count()
        return response

    def login(self):
        data = json.dumps({
            "username": self.username, "password": self.password
        })
        req = urllib2.Request(self.endpoint, data)
        req.headers['Content-Type'] = 'application/json'
        response = urllib2.urlopen(req)
        auth = json.loads(response.read())
        self.token = auth['token']
        response.close()
        return auth

    def reset_retry_count(self):
        self.retried = 0

    def retry_auth(self, req, headers):
        if self.retried >= self.max_retries:
            return urllib2.HTTPError(
                req.get_full_url(), 401, "Auth failed", headers, None
            )
        auth = self.login()
        self.reset_retry_count()
        req.headers['Cookie'] = "iPlanetDirectoryPro={}".format(auth['token'])
        return self.parent.open(req, timeout=req.timeout)


class Urllib2Executor(urllib2_executor.Urllib2Executor):
    def __init__(self, extra_handlers):
        self.handlers = extra_handlers


def use(extra_handlers=()):
    base.use_executor(Urllib2Executor(extra_handlers=extra_handlers))


from marshmallow import Serializer, fields


class CustomerAddress(Serializer):
    buildingId = fields.UUID()
    city = fields.String()
    cityCode = fields.String()
    countryCode = fields.String()
    street = fields.String()
    postalCode = fields.String()


class CustomerProfileEducationLevel(Serializer):
    edu_prim = fields.Integer()
    edu_sec = fields.Integer()
    edu_uni = fields.Integer()
    edu_noStudies = fields.Integer()


class CustomerProfile(Serializer):
    totalPersonNumber = fields.Integer()
    minorsPersonsNumber = fields.Integer()
    workingAgePersonsNumber = fields.Integer()
    retiredAgePersonsNumber = fields.Integer()
    malePersonsNumber = fields.Integer()
    femalePersonsNumber = fields.Integer()
    educationLevel = fields.Nested(CustomerProfileEducationLevel)


class Customer(Serializer):
    customerId = fields.UUID()
    address = fields.Nested(CustomerAddress)
    profile = fields.Nested(CustomerProfile)


class Device(Serializer):
    dateStart = fields.DateTime(format='iso')
    dateEnd = fields.DateTime(format='iso')
    deviceId = fields.UUID()


class Contract(Serializer):
    payerId = fields.UUID()
    ownerId = fields.UUID()
    signerId = fields.UUID()
    dateStart = fields.DateTime(format='iso')
    dateEnd = fields.DateTime(format='iso')
    contractId = fields.String()
    tariffId = fields.String()
    power = fields.Integer()
    version = fields.Integer()
    activityCode = fields.String()
    meteringPointId = fields.UUID()
    customer = fields.Nested(Customer)
    devices = fields.List(fields.Nested(Device))


class Reading(Serializer):
    type = fields.Select(['electricityConsumption'])
    unit = fields.Select(['kWh', 'Wh'])
    period = fields.Select(['INSTANT', 'CUMULATIVE', 'PULSE'])


class Measurement(Serializer):
    type = fields.Select(['electricityConsumption'])
    timestamp = fields.DateTime(format='iso')
    value = fields.Float()


class AmonMeasure(Serializer):
    deviceId = fields.UUID()
    meteringPointId = fields.UUID()
    readings = fields.List(fields.Nested(Reading))
    measurements = fields.List(fields.Nested(Measurement))



#!/usr/bin/env python
"""
Empowering
======

Python library for the Empowering REST API.

:copyright: (c) 2013 by GISCE-TI, S.L., see AUTHORS for more details.
:license: MIT, see LICENSE for more details.
"""
from setuptools import setup, find_packages

tests_require = [
    'Flask'
]

install_requires = [
    'libsaas',
    'marshmallow'
]

setup(
    name='empowering',
    version='0.13.0',
    author='GISCE-TI, S.L.',
    author_email='devel@gisce.net',
    url='http://code.gisce.net/empowering',
    description='Python library for the Empowering REST API.',
    long_description=__doc__,
    license='MIT',
    packages=find_packages(exclude=['tests', 'test']),
    zip_safe=False,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={'test': tests_require},
    test_suite='nose.collector',
    include_package_data=True,
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



from empowering.service import EmpoweringResource
from empowering.utils import searchparams_to_querystring

import calendar

class OTResult(EmpoweringResource):
    def pull(self, period=None, contract=None):
        search_params = []

        if contract:
            if type(contract) is not str:
                contract = str(contract)
            param = ('contractId', '=', contract)
            search_params.append(param)

        if period:
            if type(period) is not int:
                period = int(period)
            param = ('month', '=', period)
            search_params.append(param)

        query = searchparams_to_querystring(search_params)
        return self.multiget(where=query)

class OT101Results(OTResult):
    path = 'OT101Results'


class OT103Results(OTResult):
    path = 'OT103Results'


class OT106Results(OTResult):
    path = 'OT106Results'


class OT201Results(OTResult):
    path = 'OT201Results'


class OT204Results(OTResult):
    path = 'OT204Results'


class OT503Results(OTResult):
    path = 'OT503Results'

    def pull(self, period=None, contract=None):
        # Thanks empowering for keeping the acorded API :D </ironic>

        last_day = 31
        search_params = []
        if contract:
            if type(contract) is not str:
                contract = str(contract)
            param = ('contractId', '=', contract)
            search_params.append(param)
        if period:
            if type(period) is not str:
                # For API coherence period should be int but
                # is not useful here
                period = str(period)

            year = period[:4]
            month = period[-2:]
            last_day = calendar.monthrange(int(year), int(month))[1]

            start_period = '%s%s%s' % (year, month, '01')
            end_period = '%s%s%s' % (year, month, last_day)

            params = [
                ('day', '>=', int(start_period)),
                ('day', '<=', int(end_period)),
            ]
            search_params.extend(params)

        query = searchparams_to_querystring(search_params)
        return self.multiget(where=query, sort='[("day", 1)]',
                             max_results=last_day)


class OT900Results(OTResult):
    path = 'OT900Results'


class BT111Results(OTResult):
    path = 'BT111Results'
